#ifndef CG_H
#define CG_H 1

#include <cmath>
#include "operators.h"

//sample call: cg(A,x,b);

template <typename MATTYPE, typename VECTYPE>
int
cg(MATTYPE &A, VECTYPE &x, const VECTYPE &b, double tol = 0.0000000000001, int maxIterations = 1000000){
    double alpha, beta, rNormSquare, rNormSquarePrev;
    VECTYPE Ap, r, p;

    if ((int)x.size()!=A.numCols()) {
        x = VECTYPE(A.numCols());
    }

    r = A*x - b;
    p = -1*r;
    rNormSquare = r*r;
    for (int k=1; k<=maxIterations; k++) {
        #ifdef SOLVER_DEBUG
            std::cerr << "k = " << k << ", rho = " << sqrt(rNormSquare)
                << std::endl;
        #endif
        if (sqrt(rNormSquare)<=tol) {
            return k-1;
        }
        Ap = A*p;
        alpha = rNormSquare/(p * Ap);
        x = x + alpha*p;
        r = r + alpha*Ap;

        rNormSquarePrev = rNormSquare;
        rNormSquare = r*r;
        beta = rNormSquare/rNormSquarePrev;
        p = beta*p - r;
    }
    return maxIterations;
}


#endif // CG_H

